@@ -15,8 +15,10 @@
 #' @param operator The operator applied elementwise within the kernel, as a
 #'   function of the original image value and the kernel value. Arithmetic
 #'   operators are as usual; \code{"i"} is the identity operator, where every
-#'   value within the kernel will be included as-is; and \code{1} and \code{0}
-#'   include a 1 or 0 for each element within the kernel's nonzero region.
+#'   value within the kernel will be included as-is; \code{"1"} and \code{"0"}
+#'   include a 1 or 0 for each element within the kernel's nonzero region;
+#'   \code{"=="} produces a 1 where the image matches the kernel, and 0
+#'   elsewhere.
 #' @param merge The operator applied to combine the elements into a final value
 #'   for the centre pixel. All have their usual meanings.
 #' @param value An optional vector of values in the target array for which to
@@ -31,6 +33,10 @@
 #' @param nNeighboursNot An optional numeric vector giving nonallowable numbers
 #'   of nonzero neighbours (including diagonal neighbours) for array elements
 #'   where the kernel will be applied.
+#' @param renormalise If \code{TRUE}, the default, and \code{merge} is
+#'   \code{"sum"}, the sum will be renormalised relative to the sum over the
+#'   visited part of the kernel. This avoids low-intensity bands around the
+#'   edges of a morphed image.
 #' @param \dots Additional arguments to methods.
 #' @return A morphed array with the same dimensions as the original array.
 #' 
@@ -49,22 +55,22 @@ morph <- function (x, kernel, ...)
 
 #' @rdname morph
 #' @export
-morph.default <- function (x, kernel, operator = c("+","-","*","i","1","0"), merge = c("sum","min","max","mean","median"), value = NULL, valueNot = NULL, nNeighbours = NULL, nNeighboursNot = NULL, ...)
+morph.default <- function (x, kernel, operator = c("+","-","*","i","1","0","=="), merge = c("sum","min","max","mean","median","all","any"), value = NULL, valueNot = NULL, nNeighbours = NULL, nNeighboursNot = NULL, renormalise = TRUE, ...)
 {
     x <- as.array(x)
-    if (!is.numeric(x))
-        report(OL$Error, "Target array must be numeric")
+    if (!is.numeric(x) && !is.logical(x))
+        stop("Target array must be numeric")
     
     if (!isKernelArray(kernel))
         kernel <- kernelArray(kernel)
     
     if (any(dim(kernel) %% 2 != 1))
-        report(OL$Error, "Kernel must have odd width in all dimensions")
+        stop("Kernel must have odd width in all dimensions")
     
     if (length(dim(kernel)) < length(dim(x)))
         dim(kernel) <- c(dim(kernel), rep(1,length(dim(x))-length(dim(kernel))))
     else if (length(dim(kernel)) > length(dim(x)))
-        report(OL$Error, "Kernel has greater dimensionality than the target array")
+        stop("Kernel has greater dimensionality than the target array")
     
     operator <- match.arg(operator)
     merge <- match.arg(merge)
@@ -73,7 +79,7 @@ morph.default <- function (x, kernel, operator = c("+","-","*","i","1","0"), mer
     
     restrictions <- list(value=as.double(value), valueNot=as.double(valueNot), nNeighbours=as.integer(nNeighbours), nNeighboursNot=as.integer(nNeighboursNot))
     
-    returnValue <- .Call("morph", x, kernel, operator, merge, restrictions, PACKAGE="mmand")
+    returnValue <- .Call(C_morph, x, kernel, operator, merge, restrictions, renormalise)
     
     if (length(dim(x)) > 1)
         dim(returnValue) <- dim(x)
@@ -89,17 +95,17 @@ morph.default <- function (x, kernel, operator = c("+","-","*","i","1","0"), mer
 #' @param x An object that can be coerced to a numeric array.
 #' @return A logical value indicating whether the array is binary or not.
 #'   Binary in this case means that the array contains only one unique nonzero
-#'   value.
+#'   value, which is stored with the return value in an attribute.
 #' 
 #' @author Jon Clayden <code@@clayden.org>
 #' @export
 binary <- function (x)
 {
     x <- as.array(x)
-    if (!is.numeric(x))
-        report(OL$Error, "Array must be numeric")
+    if (!is.numeric(x) && !is.logical(x))
+        stop("Array must be numeric")
     
-    return (.Call("is_binary", x, PACKAGE="mmand"))
+    return (.Call(C_is_binary, x))
 }
 
 #' Binarise a numeric array
@@ -113,8 +119,9 @@ binary <- function (x)
 #' @author Jon Clayden <code@@clayden.org>
 #' @seealso \code{\link{morph}} for the function underlying this operation, and
 #'   \code{\link{erode}} for mathematical morphology functions.
-#' @export
-binarise <- function (x)
+#' @aliases binarize
+#' @export binarise binarize
+binarise <- binarize <- function (x)
 {
     return (morph(x, kernel=1, operator="1", valueNot=0))
 }
@@ -145,7 +152,7 @@ threshold <- function (x, level, method = c("literal","kmeans"), binarise = TRUE
 {
     method <- match.arg(method)
     if (missing(level) && method == "literal")
-        report(OL$Error, "A literal threshold level is required")
+        stop("A literal threshold level is required")
     
     if (method == "literal")
     {
@@ -159,12 +166,14 @@ threshold <- function (x, level, method = c("literal","kmeans"), binarise = TRUE
         # Drop dims so that the clustering is done by intensity only
         dims <- dim(x)
         dim(x) <- NULL
+        valid <- is.finite(x)
         
-        kmeansResult <- kmeans(x, 2)
+        kmeansResult <- kmeans(x[valid], 2)
         if (binarise)
-            x <- ifelse(kmeansResult$cluster == which.min(kmeansResult$centers), 0L, 1L)
+            x[valid] <- ifelse(kmeansResult$cluster == which.min(kmeansResult$centers), 0L, 1L)
         else
-            x <- ifelse(kmeansResult$cluster == which.min(kmeansResult$centers), 0L, x)
+            x[valid] <- ifelse(kmeansResult$cluster == which.min(kmeansResult$centers), 0L, x[valid])
+        x[!valid] <- NA
         
         dim(x) <- dims
     }
@@ -246,7 +255,7 @@ sobelFilter <- function (x, dim, axis = 0)
         dim <- length(base::dim(x))
     
     if (axis < 0 || axis > dim)
-        report(OL$Error, "The axis should be between 0 and the dimensionality of the kernel")
+        stop("The axis should be between 0 and the dimensionality of the kernel")
     
     if (axis == 0)
     {
@@ -384,3 +393,114 @@ closing <- function (x, kernel)
     
     return (erode(dilate(x, kernel), kernel))
 }
+
+#' Skeletonise a numeric array
+#' 
+#' Skeletonisation is the process of thinning a shape to a medial line or
+#' surface, and can be achieved using elementary mathematical morphology
+#' operations in a number of ways. Three methods are available through this
+#' function. They are all iterative and therefore relatively time-consuming.
+#' 
+#' The default method is Lantuéjoul's formula, a union across repeated
+#' erosions, which works in any number of dimensions and may produce
+#' reasonable results on greyscale images, but does not in general produce a
+#' connected skeleton. Beucher introduced an alternative which may produce a
+#' better result (although again the skeleton may not be connected), but this
+#' implementation of the latter algorithm only applies to binary arrays. The
+#' final method uses the so-called hit-or-miss transform, which searches for
+#' exact patterns in the source array. This is guaranteed to produce a
+#' connected skeleton, which is often desirable, but uses fixed kernels (so the
+#' \code{kernel} argument is ignored) and is currently only implemented for 2D
+#' binary arrays.
+#' 
+#' @param x An object that can be coerced to an array, or for which a
+#'   \code{\link{morph}} method exists.
+#' @param kernel An array representing the kernel to be used for the underlying
+#'   morphology operations. The kernel is fixed for the \code{"hitormiss"}
+#'   method, so this argument will be ignored.
+#' @param method A string giving the method to use. See Details.
+#' @return A skeletonised array with the same dimensions as the original array.
+#' 
+#' @examples
+#' x <- c(0,0,1,0,0,0,1,1,1,0,0)
+#' k <- c(1,1,1)
+#' skeletonise(x,k)
+#' @author Jon Clayden <code@@clayden.org>
+#' @seealso \code{\link{morphology}}
+#' @references
+#' C. Lantuéjoul (1977). Sur le modèle de Johnson-Mehl généralisé. Technical
+#' report, Centre de Morphologie Mathématique, Fontainebleau, France.
+#' 
+#' S. Beucher (1994). Digital skeletons in Euclidean and geodesic spaces.
+#' Signal Processing 38(1):127-141.
+#' \url{https://doi.org/10.1016/0165-1684(94)90061-2}.
+#' @aliases skeletonize
+#' @export skeletonise skeletonize
+skeletonise <- skeletonize <- function (x, kernel = NULL, method = c("lantuejoul","beucher","hitormiss"))
+{
+    method <- match.arg(method)
+    isBinary <- binary(x)
+    x <- result <- as.array(x)
+    nDims <- sum(dim(x) > 1)
+    
+    if (method == "lantuejoul")
+    {
+        result <- x - opening(x, kernel)
+        eroded <- x
+    
+        repeat
+        {
+            eroded <- erode(eroded, kernel)
+            if (all(eroded == 0))
+                break
+            result <- pmax(result, eroded - opening(eroded,kernel))
+        }
+    }
+    else if (method == "beucher")
+    {
+        if (!isBinary)
+            warning("The Beucher skeletonisation method will not preserve a greyscale")
+        
+        repeat
+        {
+            result <- x
+            x <- erode(x,kernel) | (dilate(x - opening(x,kernel), kernel) & x)
+            if (isTRUE(all.equal(x, result)))
+                break
+        }
+    }
+    else
+    {
+        if (!isBinary)
+            stop("The hit-or-miss transform skeletonisation method only works with binary images")
+        if (nDims != 2)
+            stop("The hit-or-miss transform skeletonisation method is only implemented in 2D")
+        
+        x <- x / attr(isBinary, "value")
+        k1 <- matrix(c(0,NA,1,0,1,1,0,NA,1), 3, 3)
+        k2 <- matrix(c(NA,1,NA,0,1,1,0,0,NA), 3, 3)
+        rotateKernel <- function(k) t(apply(k, 2, rev))
+        hitOrMiss <- function(x,k) morph(x, k, operator="==", merge="all", value=1)
+    
+        repeat
+        {
+            result <- x
+            for (i in 1:4)
+            {
+                x <- x & !hitOrMiss(x, k1)
+                x <- x & !hitOrMiss(x, k2)
+                k1 <- rotateKernel(k1)
+                k2 <- rotateKernel(k2)
+            }
+        
+            if (isTRUE(all.equal(x, result)))
+                break
+        }
+    }
+    
+    result <- as.double(result)
+    if (length(dim(x)) > 1)
+        dim(result) <- dim(x)
+    
+    return (result)
+}
